package ru.dbotthepony.mc.otm.core

import io.netty.handler.codec.EncoderException
import net.minecraft.nbt.CompoundTag
import net.minecraft.nbt.NbtAccounter
import net.minecraft.nbt.NbtIo
import net.minecraft.world.item.Item
import net.minecraft.world.item.ItemStack
import net.minecraftforge.registries.ForgeRegistries
import net.minecraftforge.registries.ForgeRegistry
import java.io.*
import java.math.BigDecimal
import java.math.BigInteger

// But seriously, Mojang, why would you need to derive from ByteBuf directly, when you can implement
// your own InputStream and OutputStream, since ByteBuf is meant to be operated on most time like a stream anyway?

// netty ByteBuf -> netty ByteBufInputStream -> Minecraft FriendlyInputStream

fun OutputStream.writeInt(value: Int) {
	if (this is DataOutputStream) {
		writeInt(value)
		return
	}

	write(value ushr 24)
	write(value ushr 16)
	write(value ushr 8)
	write(value)
}

fun InputStream.readInt(): Int {
	if (this is DataInputStream) {
		return readInt()
	}

	return (read() shl 24) or (read() shl 16) or (read() shl 8) or read()
}

fun OutputStream.writeLong(value: Long) {
	if (this is DataOutputStream) {
		writeLong(value)
		return
	}

	write((value ushr 48).toInt())
	write((value ushr 40).toInt())
	write((value ushr 32).toInt())
	write((value ushr 24).toInt())
	write((value ushr 16).toInt())
	write((value ushr 8).toInt())
	write(value.toInt())
}

fun InputStream.readLong(): Long {
	if (this is DataInputStream) {
		return readLong()
	}

	return (read().toLong() shl 48) or
		(read().toLong() shl 40) or
		(read().toLong() shl 32) or
		(read().toLong() shl 24) or
		(read().toLong() shl 16) or
		(read().toLong() shl 8) or
		read().toLong()
}

fun OutputStream.writeFloat(value: Float) = writeInt(value.toBits())
fun InputStream.readFloat() = Float.fromBits(readInt())
fun OutputStream.writeDouble(value: Double) = writeLong(value.toBits())
fun InputStream.readDouble() = Double.fromBits(readLong())

fun OutputStream.writeNbt(value: CompoundTag) {
	try {
		NbtIo.write(value, if (this is DataOutputStream) this else DataOutputStream(this))
	} catch (ioexception: IOException) {
		throw EncoderException(ioexception)
	}
}

fun InputStream.readNbt(accounter: NbtAccounter = NbtAccounter.UNLIMITED): CompoundTag {
	return try {
		NbtIo.read(if (this is DataInputStream) this else DataInputStream(this), accounter)
	} catch (ioexception: IOException) {
		throw EncoderException(ioexception)
	}
}

fun OutputStream.writeItem(itemStack: ItemStack, limitedTag: Boolean = true) {
	if (itemStack.isEmpty) {
		write(0)
	} else {
		write(1)
		val id = (ForgeRegistries.ITEMS as ForgeRegistry<Item>).getID(itemStack.item)

		writeInt(id)
		writeInt(itemStack.count)

		var compoundtag: CompoundTag? = null

		if (itemStack.item.isDamageable(itemStack) || itemStack.item.shouldOverrideMultiplayerNbt()) {
			compoundtag = if (limitedTag) itemStack.shareTag else itemStack.tag
		}

		write(if (compoundtag != null) 1 else 0)

		if (compoundtag != null) {
			writeNbt(compoundtag)
		}
	}
}

fun InputStream.readItem(): ItemStack {
	if (read() == 0) {
		return ItemStack.EMPTY
	}


	val item = (ForgeRegistries.ITEMS as ForgeRegistry<Item>).getValue(readInt())
	val itemStack = ItemStack(item, readInt())

	if (read() != 0) {
		itemStack.readShareTag(readNbt())
	}

	return itemStack
}

fun InputStream.readVarIntBE(): Int {
	var result = 0
	var read = read()

	do {
		result = (result shl 7) or (read and 127)
		read = read()
	} while (read and 128 != 0)

	return result
}

fun InputStream.readVarIntLE(): Int {
	var result = 0
	var read = read()
	var i = 0

	while (read and 128 != 0) {
		result = result or ((read and 127) shl i)
		read = read()
		i += 7
	}

	result = result or ((read and 127) shl i)

	return result
}

fun OutputStream.writeVarIntLE(value: Int) {
	require(value >= 0) { "Negative number provided: $value" }
	var written = value

	while (written >= 128) {
		write((written and 127) or 128)
		written = written ushr 7
	}

	write(written)
}

fun OutputStream.writeBigDecimal(value: BigDecimal) {
	writeInt(value.scale())
	val bytes = value.unscaledValue().toByteArray()
	writeVarIntLE(bytes.size)
	write(bytes)
}

fun InputStream.readBigDecimal(): BigDecimal {
	val scale = readInt()
	val size = readVarIntLE()
	val bytes = ByteArray(size)
	read(bytes)
	return BigDecimal(BigInteger(bytes), scale)
}
